﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using SimpleJSON;

namespace VAMLivePlugin
{
    public static class Utils
    {
	    public static Transform GetChildNamed( Transform parent, string childName ) {
		    foreach( Transform child in parent )
		    {
			    if( child.name == childName ) {
				    return child;
			    } else {
				    Transform childSearch = GetChildNamed(child,childName);
				    if (childSearch != null) return childSearch;
			    }
		    }
        
		    return null;
	    }
	    
	    public static UIDynamicTextField CreateStaticDescriptionText(MVRScript script, string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft, bool disableBackground = true, bool disableScroll = true ) {
		    JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
		    staticDescString.hidden = true;
		    UIDynamicTextField staticDescStringField = script.CreateTextField(staticDescString, rightSide);
		    if( disableBackground ) staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
		    staticDescStringField.UItext.alignment = textAlignment;
		    LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
		    sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
		    staticDescStringField.height = fieldHeight;
		    if( disableScroll ) Utils.DisableScrollOnText(staticDescStringField);
		    return staticDescStringField;
	    }
	    
	    public static void SetupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = false, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) DisableScrollOnText(target);
		}
	    
	    public static void DisableScrollOnText(UIDynamicTextField target) {
		    ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
		    if( targetSR != null ) {
			    targetSR.horizontal = false;
			    targetSR.vertical = false;
		    }
	    }
	    
	    public static void ScrollToTop(UIDynamicTextField target) {
		    ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
		    if( targetSR != null ) {
			    targetSR.normalizedPosition = new Vector2(0, 1);
		    }
	    }

	    public static void SetupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

	    public static void SetupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
	    public static void SetupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

	    public static void SetButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
	    public static void SetButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
	    public static void UpdateButtonPositionWithoutLayout(UIDynamicButton target, Vector2 newPosition)
		{
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			tmpRectT.anchoredPosition = newPosition;
		}
		
		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}	
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Breathing Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
		
	}
}
