using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// **********************************
// VAMLive Screen 1.0
// **********************************

namespace VAMLivePlugin
{
    public partial class VAMLiveScreen : MVRScript
    {	
	    // Static string
        public static string PLUGIN_PATH;
        private static string CAM_BUNDLE_PATH = "/vamlive_assets.cambundle";
	    private static string MATERIALS_PATH = "assets/vamlive/materials/screen/";
	    private static string MASK_PATH = "assets/vamlive/textures/masks/";
	    private static string MASKALPHA_PATH = "assets/vamlive/textures/masksalpha/";
        public static string INFO_PREFIX = "VAMLive: ";
        public static string EXCEPTION_PREFIX = "VAMLive - Exception caught: ";
        
        // JSONStorables
        protected JSONStorableStringChooser ScreenOpt_SourceCam;
        protected List<string> _sourceCameChoices = new List<string> {"None"};

        protected JSONStorableFloat ScreenOpt_ScreenSizeX;
        protected JSONStorableFloat ScreenOpt_ScreenRatioWidth;
        protected JSONStorableFloat ScreenOpt_ScreenRatioHeight;
        protected JSONStorableFloat ScreenOpt_Tiling;

        protected JSONStorableStringChooser ScreenOpt_Material;
        protected List<string> _materialChoices = new List<string>();
        
        protected JSONStorableColor ScreenOpt_ScreenColor;
        protected JSONStorableFloat ScreenOpt_ScreenBrightness;
        protected JSONStorableFloat ScreenOpt_ScreenContrast;
        protected JSONStorableFloat ScreenOpt_ScreenGrayscale;

        protected JSONStorableStringChooser ScreenOpt_MaskType;
        protected List<string> _masksChoices = new List<string> {"circle_01"};

        protected JSONStorableFloat ScreenOpt_MaskSizeX;
        protected JSONStorableFloat ScreenOpt_MaskSizeY;
        protected JSONStorableColor ScreenOpt_MaskColor;
        
        // The usual help section
        protected JSONStorableString helpText;
        
        // Global UI components
        protected UIDynamicTextField helpWindow;
        
        // Plugins
        private JSONStorable VAMLiveCamPlugin;
        
        // VAMS External values
        public JSONStorableBool VAMLValue_Ready;
        
        // Game Objects, components, assets and assets lists
        protected GameObject VAMLIVESCREENGO;
        
        protected Transform _TRVAMLScreenSurface;
        protected Renderer _RENDScreenSurface;
        protected Material _MATScreenSurface;
        
        protected Transform _TRVAMLSourceCamera;
        protected Camera _VAMLSourceCamera;

        protected List<Material> _materialList = new List<Material>();
        protected List<Texture2D> _masksList = new List<Texture2D>();
        protected List<Texture2D> _masksAlphasList = new List<Texture2D>();
        
        // Textures
        protected Texture2D _maskTex;
        
        // Dynamic UI
        protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
        
        // Miscs
        protected float _vamexmTimer = 0f;
        protected bool _vamliveInitialized = false;
        protected bool _isBundleLoaded = false;
        protected bool _cameraReady = false;
        protected float _cameraReadyCheckPeriod = 0.3f;
        
        // Assetbundle
        private Request bundleRequest = null;
        
        // Security
        protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
        protected bool _isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }

		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError(EXCEPTION_PREFIX + "Please add VAMLive Screen to an Empty atom");
                    return;
                }
                
                PLUGIN_PATH = Utils.GetPluginPath(this);

                /* *** Temporary UIDynamic *** */
                HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                HSVColor hsvcDefaultBlack = HSVColorPicker.RGBToHSV(0f, 0f, 0f);
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamicColorPicker tmpColor;
                UIDynamic tmpSpacer;
                
                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "<b>VAMLIVE SCREEN HELP:</b>\n" +
	                "This plugin adds a screen that can display a camera from VAMLive.\n\n" +
	                "<b>CAMERA:</b>\n" +
	                "Select a source with the list in the right column.\n\n" +
	                "You need to have a camera already added to the scene.\n\n" +
	                "<b>SCREEN:</b>\n" +
	                "You can only modify the with of the screen, the height will be computed based on the ratio. " +
	                "You can click on <b>Use source camera ratio</b> to copy the ratio from your source camera.\n\n" +
	                "To avoid having a distorted image, you need to have the same ratio on your camera and screen.\n\n" +
	                "<b>MATERIAL / MASK:</b>\n" +
	                "These are cosmetic options, you can tweak them to your liking. Mask options are only active with materials containing the keyword <b>masked</b>." +
	                "</color></size>"
                );
				
                helpWindow = CreateTextField(helpText, false);
                helpWindow.height = 1150.0f;
                
                /* ********************************* */
                /* **** RIGHT UI                **** */
                /* ********************************* */
                Utils.CreateStaticDescriptionText(this,"VAMLiveScreen","<color=#000><size=35><b>VAMLive (Screen)</b></size></color>",true,40,TextAnchor.MiddleRight);

                // **** Cam options ****
                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Cam","<color=#000><size=28><b>CAMERA</b></size></color>\n<size=28>Source camera options for the screen.</size>",true,65,TextAnchor.MiddleRight);

                ScreenOpt_SourceCam = new JSONStorableStringChooser("ScreenOpt_SourceCam", _sourceCameChoices, _sourceCameChoices.First(), "Source");
                ScreenOpt_SourceCam.setCallbackFunction += (val) => { OnChangeSourceCamera(); };
                tmpPopup = CreateScrollablePopup(ScreenOpt_SourceCam, true);
                tmpPopup.labelWidth = 125f;

                // **** Screen size options ****
                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);

                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Screen","<color=#000><size=28><b>SCREEN</b></size></color>",true,35,TextAnchor.MiddleRight);

                ScreenOpt_ScreenSizeX = new JSONStorableFloat("ScreenOpt_ScreenSizeX", 0.25f, 0.001f, 2f);
                ScreenOpt_ScreenSizeX.setCallbackFunction += (val) => { OnChangeScreenSize(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenSizeX, true);
                tmpSlider.label = "Screen width";

                ScreenOpt_ScreenRatioWidth = new JSONStorableFloat("ScreenOpt_ScreenRatioWidth", 16f, 1f, 32f);
                ScreenOpt_ScreenRatioWidth.setCallbackFunction += (val) => { ScreenOpt_ScreenRatioWidth.val = Mathf.Round(val); OnChangeScreenSize(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenRatioWidth, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Ratio (width)";

                ScreenOpt_ScreenRatioHeight = new JSONStorableFloat("ScreenOpt_ScreenRatioHeight", 9f, 1f, 9f);
                ScreenOpt_ScreenRatioHeight.setCallbackFunction += (val) => { ScreenOpt_ScreenRatioHeight.val = Mathf.Round(val); OnChangeScreenSize(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenRatioHeight, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Ratio (height)";
                
                tmpButton = CreateButton("Use source camera ratio", true);
                tmpButton.button.onClick.AddListener(() => { OnUseCameraRatio(); } );
                Utils.SetButtonColor(tmpButton, new Color(0.1f,0.2f,0.5f,1f));
                Utils.SetButtonTextColor(tmpButton, new Color(1f,1f,1f,1f));
                
                ScreenOpt_Tiling = new JSONStorableFloat("ScreenOpt_Tiling", 1f, 1f, 20f);
                ScreenOpt_Tiling.setCallbackFunction += (val) => { ScreenOpt_Tiling.valNoCallback = Mathf.Round(val); OnChangeScreenSize(); };
                tmpSlider = CreateSlider(ScreenOpt_Tiling, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Tiling";

                // **** Material options ****
                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);

                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Material","<color=#000><size=28><b>MATERIAL</b></size></color>",true,35,TextAnchor.MiddleRight);

                ScreenOpt_Material = new JSONStorableStringChooser("ScreenOpt_Material", _materialChoices, "screen", "Material");
                ScreenOpt_Material.setCallbackFunction += (val) => { OnChangeMaterial(); };
                tmpPopup = CreateScrollablePopup(ScreenOpt_Material, true);
                tmpPopup.labelWidth = 125f;

                ScreenOpt_ScreenColor = new JSONStorableColor("ScreenOpt_ScreenColor", hsvcDefault);
                ScreenOpt_ScreenColor.setJSONCallbackFunction += (val) => { OnChangeScreenColor(); };
                tmpColor = CreateColorPicker(ScreenOpt_ScreenColor, true);
                tmpColor.label = "Screen color";
                
                ScreenOpt_ScreenBrightness = new JSONStorableFloat("ScreenOpt_ScreenBrightness", 1f, 0f, 10f, false);
                ScreenOpt_ScreenBrightness.setCallbackFunction += (val) => { OnChangeScreenBrightness(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenBrightness, true);
                tmpSlider.label = "Screen Brightness";
                tmpSlider.valueFormat = "F2";
                
                ScreenOpt_ScreenContrast = new JSONStorableFloat("ScreenOpt_ScreenContrast", 1f, 0f, 10f, false);
                ScreenOpt_ScreenContrast.setCallbackFunction += (val) => { OnChangeScreenContrast(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenContrast, true);
                tmpSlider.label = "Screen Contrast";
                tmpSlider.valueFormat = "F2";
                
                ScreenOpt_ScreenGrayscale = new JSONStorableFloat("ScreenOpt_ScreenGrayscale", 0f, 0f, 1f);
                ScreenOpt_ScreenGrayscale.setCallbackFunction += (val) => { OnChangeScreenGrayscale(); };
                tmpSlider = CreateSlider(ScreenOpt_ScreenGrayscale, true);
                tmpSlider.label = "Screen Grayscale";
                tmpSlider.valueFormat = "F2";
                
                // **** Mask options ****
                tmpSpacer = CreateSpacer(true);
                Utils.SetupSpacer(tmpSpacer, 20f);

                tmpTextfield = Utils.CreateStaticDescriptionText(this,"Mask","<color=#000><size=28><b>MASK</b></size></color>",true,35,TextAnchor.MiddleRight);
                
                ScreenOpt_MaskType = new JSONStorableStringChooser("ScreenOpt_MaskType", _masksChoices, _masksChoices.First(), "Type");
                ScreenOpt_MaskType.setCallbackFunction += (val) => { OnChangeMaskType(); };
                tmpPopup = CreateScrollablePopup(ScreenOpt_MaskType, true);
                tmpPopup.labelWidth = 125f;
                
                ScreenOpt_MaskSizeX = new JSONStorableFloat("ScreenOpt_MaskSizeX", 400f, 1f, 3000f);
                ScreenOpt_MaskSizeX.setCallbackFunction += (val) => { OnChangeMaskSize(); };
                tmpSlider = CreateSlider(ScreenOpt_MaskSizeX, true);
                tmpSlider.label = "Mask Size X";
                tmpSlider.valueFormat = "F0";

                ScreenOpt_MaskSizeY = new JSONStorableFloat("ScreenOpt_MaskSizeY", 300f, 1f, 3000f);
                ScreenOpt_MaskSizeY.setCallbackFunction += (val) => { OnChangeMaskSize(); };
                tmpSlider = CreateSlider(ScreenOpt_MaskSizeY, true);
                tmpSlider.label = "Mask Size Y";
                tmpSlider.valueFormat = "F0";
                
                ScreenOpt_MaskColor = new JSONStorableColor("ScreenOpt_MaskColor", hsvcDefaultBlack);
                ScreenOpt_MaskColor.setJSONCallbackFunction += (val) => { OnChangeMaskColor(); };
                tmpColor = CreateColorPicker(ScreenOpt_MaskColor, true);
                tmpColor.label = "Mask color";
                
                // ******* CREATING MAIN TRIGGERS/ACTIONS/EXTERNAL VALUES ********
                // They are here to be displayed on TOP of every other JSONStorables
                VAMLValue_Ready = new JSONStorableBool("VAMLValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMLValue_Ready);
                
                /* ***************************************** */
                /* **** REGISTERING STORABLES/ACTIONS   **** */
                /* ***************************************** */

                RegisterStringChooser(ScreenOpt_SourceCam);
                
                RegisterFloat(ScreenOpt_ScreenSizeX);
                RegisterFloat(ScreenOpt_ScreenRatioWidth);
                RegisterFloat(ScreenOpt_ScreenRatioHeight);
                RegisterFloat(ScreenOpt_Tiling);

                RegisterStringChooser(ScreenOpt_Material);
                RegisterColor(ScreenOpt_ScreenColor);
                RegisterFloat(ScreenOpt_ScreenBrightness);
                RegisterFloat(ScreenOpt_ScreenContrast);
                RegisterFloat(ScreenOpt_ScreenGrayscale);
                
                RegisterStringChooser(ScreenOpt_MaskType);
                RegisterFloat(ScreenOpt_MaskSizeX);
                RegisterFloat(ScreenOpt_MaskSizeY);
                RegisterColor(ScreenOpt_MaskColor);


                // Loading the assetbundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + CAM_BUNDLE_PATH, callback = OnBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Update()
		{
			if (_isSceneLoading || _isSceneFrozen) return;

			if (!_isSceneLoading && _isBundleLoaded && _vamliveInitialized == false)
			{
				if (ScreenOpt_SourceCam.val != "None" && _vamliveInitialized == false && VAMLValue_Ready.val == false )
				{
					StartCoroutine("WaitForCamera");
					VAMLValue_Ready.val = true;
				}
				else if(_vamliveInitialized == false && _cameraReady == true)
				{
					_TRVAMLScreenSurface = Utils.GetChildNamed(VAMLIVESCREENGO.transform, "screen_surface");
					_RENDScreenSurface = _TRVAMLScreenSurface.GetComponent<Renderer>();
					_MATScreenSurface = _RENDScreenSurface.material;

					_vamliveInitialized = true;
				
					// Initializing object
					OnChangeSourceCamera();
					OnChangeMaterial();
					OnChangeScreenColor();
					OnChangeScreenBrightness();
					OnChangeScreenContrast();
					OnChangeScreenGrayscale();
					OnChangeMaskSize();
					OnChangeMaskColor();
					OnChangeScreenSize();					
				}
			}
			
			// No update if not ready
			if (!_vamliveInitialized) return;

			// Internal timer
			_vamexmTimer += Time.deltaTime;

		}

		// **************************
		// Functions
		// **************************

		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		protected void OnChangeSourceCamera()
		{
			if (!_vamliveInitialized) return;
			if (ScreenOpt_SourceCam.val == "None") return;

			Atom tmpAtom = GetAtomById(ScreenOpt_SourceCam.val);
			
			// Grabbing the camera
			_TRVAMLSourceCamera = Utils.GetChildNamed(tmpAtom.reParentObject, "screen_camera");
			if( _TRVAMLSourceCamera != null ) {
				_VAMLSourceCamera = _TRVAMLSourceCamera.GetComponent<Camera>();
				
				_MATScreenSurface.SetTexture("_MainTex", _VAMLSourceCamera.targetTexture ) ;
				
				foreach (string stId in tmpAtom.GetStorableIDs())
				{
					if (stId.Contains(".VAMLiveCamera"))
					{
						VAMLiveCamPlugin = tmpAtom.GetStorableByID(stId);
					}
				}
			}
			
		}

		protected void OnUseCameraRatio()
		{
			if (!_vamliveInitialized) return;
			if (VAMLiveCamPlugin != null)
			{
				ScreenOpt_ScreenRatioWidth.val = VAMLiveCamPlugin.GetFloatParamValue("CamOpt_RatioWidth");
				ScreenOpt_ScreenRatioHeight.val = VAMLiveCamPlugin.GetFloatParamValue("CamOpt_RatioHeight");
			}
		}

		protected void OnChangeMaterial() {
			if (!_vamliveInitialized) return;
			if (_VAMLSourceCamera == null) return;

			_RENDScreenSurface.material = _materialList[_materialChoices.IndexOf(ScreenOpt_Material.val)]; 
			_MATScreenSurface = _RENDScreenSurface.material;
			
			// Updating material
			_MATScreenSurface.SetTexture("_MainTex", _VAMLSourceCamera.targetTexture ) ;
			OnChangeScreenColor();
			OnChangeScreenBrightness();
			OnChangeScreenContrast();
			OnChangeScreenGrayscale();
			OnChangeMaskType();
			OnChangeMaskSize();
			OnChangeMaskColor();
		}

		protected void OnChangeScreenColor()
		{
			if (!_vamliveInitialized) return;
			
			Color newScreenColor = new Color(ScreenOpt_ScreenColor.colorPicker.currentColor.r, ScreenOpt_ScreenColor.colorPicker.currentColor.g, ScreenOpt_ScreenColor.colorPicker.currentColor.b, 1f);
			_MATScreenSurface.SetColor("_ScreenColor", newScreenColor);
		}
		
		protected void OnChangeScreenBrightness()
		{
			if (!_vamliveInitialized) return;
			
			_MATScreenSurface.SetFloat("_Brightness", ScreenOpt_ScreenBrightness.val);
		}
		
		protected void OnChangeScreenContrast()
		{
			if (!_vamliveInitialized) return;
			
			_MATScreenSurface.SetFloat("_Contrast", ScreenOpt_ScreenContrast.val);
		}
		
		protected void OnChangeScreenGrayscale()
		{
			if (!_vamliveInitialized) return;
			
			_MATScreenSurface.SetFloat("_GrayscaleAmount", ScreenOpt_ScreenGrayscale.val);
		}

		protected void OnChangeMaskType()
		{
			if (!_vamliveInitialized) return;

			if (ScreenOpt_Material.val == "screenmasked")
			{
				_MATScreenSurface.SetTexture("_Mask", _masksList[_masksChoices.IndexOf(ScreenOpt_MaskType.val)] );				
			}
			else if (ScreenOpt_Material.val == "screenmaskedtransparent")
			{
				_MATScreenSurface.SetTexture("_Mask", _masksAlphasList[_masksChoices.IndexOf(ScreenOpt_MaskType.val)] );
			}
		}
		
		protected void OnChangeMaskSize()
		{
			if (!_vamliveInitialized) return;
			_MATScreenSurface.SetTextureScale("_Mask", new Vector2(ScreenOpt_MaskSizeX.val, ScreenOpt_MaskSizeY.val));
		}

		protected void OnChangeMaskColor()
		{
			if (!_vamliveInitialized) return;
			
			Color newMaskColor = new Color(ScreenOpt_MaskColor.colorPicker.currentColor.r, ScreenOpt_MaskColor.colorPicker.currentColor.g, ScreenOpt_MaskColor.colorPicker.currentColor.b, 1f);
			_MATScreenSurface.SetColor("_MaskColor", newMaskColor);
		}

		protected void OnChangeScreenSize()
		{
			if (!_vamliveInitialized) return;

			_MATScreenSurface.SetTextureScale("_MainTex", new Vector2(ScreenOpt_Tiling.val, ScreenOpt_Tiling.val));
			float screenWidth = ScreenOpt_ScreenSizeX.val;
			float screenHeight = (ScreenOpt_ScreenRatioHeight.val / ScreenOpt_ScreenRatioWidth.val) * screenWidth;
			_TRVAMLScreenSurface.localScale = new Vector3(screenWidth, 1f, screenHeight);
		}
		
		protected void OnBundleLoaded(Request aRequest) {
			try
			{
				// Reading my UI bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("assets/vamlive/prefabs/vamlive_screen.prefab");
				VAMLIVESCREENGO = Instantiate(tmp);
				
				_maskTex = bundleRequest.assetBundle.LoadAsset<Texture2D>("assets/vamlive/textures/filters/mask-01.png");
				
				// Parenting to my current Empty
				VAMLIVESCREENGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMLIVESCREENGO.transform.localPosition = new Vector3(0, 0, 0);
				VAMLIVESCREENGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
				
				_materialChoices = new List<string>();
				_masksChoices = new List<string>();
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(MATERIALS_PATH))
					{
						_materialChoices.Add(bundleFileItem.Replace(MATERIALS_PATH, "").Replace(".mat", ""));
						_materialList.Add(bundleRequest.assetBundle.LoadAsset<Material>(bundleFileItem));
					}
					
					if (bundleFileItem.Contains(MASK_PATH))
					{
						_masksChoices.Add(bundleFileItem.Replace(MASK_PATH, "").Replace(".png", ""));
						_masksList.Add(bundleRequest.assetBundle.LoadAsset<Texture2D>(bundleFileItem));
					}
					
					if (bundleFileItem.Contains(MASKALPHA_PATH))
					{
						_masksAlphasList.Add(bundleRequest.assetBundle.LoadAsset<Texture2D>(bundleFileItem));
					}
				}

				ScreenOpt_Material.choices = _materialChoices;
				ScreenOpt_MaskType.choices = _masksChoices;
				
				RefreshCameras();
				
				if (ScreenOpt_SourceCam.val == "None" && _sourceCameChoices.Count > 1)
				{
					ScreenOpt_SourceCam.val = _sourceCameChoices[1];
				}
				
				_isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
			}
		}
		
		// **************************
		// COROUTINES
		// **************************
		protected IEnumerator WaitForCamera()
		{
			// Checking VAMLive target camera
			while (_cameraReady == false)
			{
				// We're loading...
				// The reason behind that is that, a plugin looking for other atoms can load fast enough
				// to catch the storables list in the target atom PRIOR to the plugins storables being even present in the List 
				// which would result in a "false positive" of an Empty atom not owning a VAMS plugin
				if (_isSceneLoading)
				{
					yield return new WaitForSeconds(_cameraReadyCheckPeriod);					
				}
				else
				{
					bool atReady = true;
					Atom tmpAtom = GetAtomById(ScreenOpt_SourceCam.val);
					foreach (string stId in tmpAtom.GetStorableIDs())
					{
						if (stId.Contains(".VAMLiveCamera"))
						{
							JSONStorable tmpCAM = tmpAtom.GetStorableByID(stId);
							if (tmpCAM != null && tmpCAM.GetBoolParamValue("VAMLValue_Ready") == false) atReady = false;
						}
					}

					if (atReady == false)
					{
						yield return new WaitForSeconds(_cameraReadyCheckPeriod);					
					}
					else
					{
						_cameraReady = true;
						yield break;
					}
				}
			}
		}
		
		// **************************
		// BROADCASTS SEND/RECEIVE
		// **************************
		protected void RefreshCameras()
		{
			_sourceCameChoices = new List<string> {"None"};
			foreach (Atom at in SuperController.singleton.GetAtoms())
			{
				if (at.type == "Empty")
				{
					foreach (string stId in at.GetStorableIDs())
					{
						if (stId.Contains(".VAMLiveCamera"))
						{
							_sourceCameChoices.Add(at.name);
							if(ScreenOpt_SourceCam.val == at.name) VAMLiveCamPlugin = at.GetStorableByID(stId);
						}
					}
				}
			}
			ScreenOpt_SourceCam.choices = _sourceCameChoices;
		}
	
		// **************************
		// LOCAL TOOLS
		// **************************
		private static void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private static void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		protected void ToggleUIDynamicGrp(List<UIDynamic> uiGrp,bool enabled)
		{
			foreach( UIDynamic uidElem in uiGrp )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void SetupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void SetupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
			ignoreLayoutUI.Add((UIDynamic) target);
		}

		// **************************
		// Time to cleanup !
		// **************************	
		void OnDisable() {

		}

		void OnDestroy()
		{
			// Cleaning GameObjects
			Destroy(VAMLIVESCREENGO);
			
			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + CAM_BUNDLE_PATH);
		}
    }
}